#include "testLinkedList.h"
#include <malloc.h>
#include <string.h>

CPPUNIT_TEST_SUITE_REGISTRATION(LinkedListTestCase);

void LinkedListTestCase::setupBasicList()
{
  ll_add(&myList, "CS3841", 7);
  ll_add(&myList, "Operating", 11);
  ll_add(&myList, "Systems", 8);
  ll_add(&myList, "Fall", 5);
  ll_add(&myList, "Quarter", 8);
}

void LinkedListTestCase::setUp()
{
  printf(" Setup being called. ");
  // Initialize the linked list to be an empty list.
  ll_init(&myList);
}

void LinkedListTestCase::tearDown()
{
  printf(" Teardown being called. ");
  // Clear the linked list, making certain everything is empty.
  ll_clear(&myList);
}

void LinkedListTestCase::testInit()
{
  ll_init(&myList);
  CPPUNIT_ASSERT_MESSAGE("Size not zero.", ll_size(&myList)==0);
  CPPUNIT_ASSERT_MESSAGE("Head not null.", myList.head==NULL);
  CPPUNIT_ASSERT_MESSAGE("Tail not null.", myList.tail==NULL);
}

void LinkedListTestCase::testAddNull()
{
  // Check that null is handled properly, both for the list, the object, and the size.
  CPPUNIT_ASSERT_MESSAGE("Improper handling of null list", ll_add(NULL, "Hello", 6)==false);
  CPPUNIT_ASSERT_MESSAGE("Improper handling of NULL object.", ll_add(&myList, NULL, 0)==false);
  CPPUNIT_ASSERT_MESSAGE("Improper handling of 0 size.", ll_add(&myList, "Hello", 0)==false);
}

void LinkedListTestCase::testAdd()
{
  // Add a single valid element.
  CPPUNIT_ASSERT_MESSAGE("Inappropriate return value from simple add.", ll_add(&myList, "Hello", 6)==true);
  CPPUNIT_ASSERT_MESSAGE("Head and tail of list match.", myList.head==myList.tail);
  CPPUNIT_ASSERT_MESSAGE("Size is incorrect.", myList.size==1);
  
  // Add a second valid element.
  CPPUNIT_ASSERT_MESSAGE("Inappropriate return value from simple add.", ll_add(&myList, "World", 6)==true);
  CPPUNIT_ASSERT_MESSAGE("Head and tail of list match.", myList.head!=myList.tail);
  CPPUNIT_ASSERT_MESSAGE("Size is incorrect.", myList.size==2);
  CPPUNIT_ASSERT_MESSAGE("List data does not match properly for head element.", strcmp((const char *)myList.head->data,"Hello")==0);
  CPPUNIT_ASSERT_MESSAGE("List data does not match properly for tail element.", strcmp((const char *)myList.tail->data, "World")==0);
  CPPUNIT_ASSERT_MESSAGE("List does not walk properly in forward direction.", strcmp((const char *)myList.head->nextNode->data, "World")==0);
  CPPUNIT_ASSERT_MESSAGE("List does not walk properly in a backward direction.", strcmp((const char *)myList.tail->prevNode->data, "Hello")==0);

  // Add a third valid element
  CPPUNIT_ASSERT_MESSAGE("Inappropriate return value from simple add.", ll_add(&myList, "This is London.", 16)==true);
  CPPUNIT_ASSERT_MESSAGE("Head and tail of list match.", myList.head!=myList.tail);
  CPPUNIT_ASSERT_MESSAGE("Size is incorrect.", myList.size==3);
  CPPUNIT_ASSERT_MESSAGE("List data does not match properly for head element.", strcmp((const char *)myList.head->data,"Hello")==0);
  CPPUNIT_ASSERT_MESSAGE("List data does not match properly for tail element.", strcmp((const char *)myList.tail->data, "This is London.")==0);
  CPPUNIT_ASSERT_MESSAGE("List does not walk properly in forward direction.", strcmp((const char *)myList.head->nextNode->data, "World")==0);
  CPPUNIT_ASSERT_MESSAGE("List does not walk properly in forward direction.", strcmp((const char *)myList.head->nextNode->nextNode->data, "This is London.")==0);
  CPPUNIT_ASSERT_MESSAGE("List does not walk properly in a backward direction.", strcmp((const char *)myList.tail->prevNode->data, "World")==0);
  CPPUNIT_ASSERT_MESSAGE("List does not walk properly in a backward direction.", strcmp((const char *)myList.tail->prevNode->prevNode->data, "Hello")==0);
}

void LinkedListTestCase::testAddIndex()
{
  // Start by setting up a basic list.
  ll_add(&myList, "Hello", 6);
  ll_add(&myList, "This is London.", 16);

  // Check to make sure the list is in the right shape.  This should never fail unless one of the test above is also failing.
  CPPUNIT_ASSERT_MESSAGE("Size is incorrect.", myList.size==2);

  // Check to make ceratin all of the invalid parameters fail properly.
  CPPUNIT_ASSERT_MESSAGE("Invalid list parameter", ll_addIndex(NULL, "World", 6, 1)==false);
  CPPUNIT_ASSERT_MESSAGE("Invalid object to add.", ll_addIndex(&myList, NULL, 6, 1)==false);
  CPPUNIT_ASSERT_MESSAGE("Invalid size of object to add.", ll_addIndex(&myList, "World", 0, 1)==false);
  CPPUNIT_ASSERT_MESSAGE("Invalid index for add.", ll_addIndex(&myList, "World", 6, 3)==false);

  // Check to make certain adding to the tail effectivly works.
  CPPUNIT_ASSERT_MESSAGE("Improper tail element with addIndex function.", ll_addIndex(&myList, "Here is the news.", 18, 2)==true);
  CPPUNIT_ASSERT_MESSAGE("Size improper after add.", myList.size==3);
  CPPUNIT_ASSERT_MESSAGE("Tail set up properly.", strcmp("Here is the news.", (const char*)myList.tail->data)==0);
  CPPUNIT_ASSERT_MESSAGE("Tail not given null as next element.", myList.tail->nextNode==NULL);
  CPPUNIT_ASSERT_MESSAGE("Previous not set up properly from tail add.", strcmp("This is London.", (const char*)myList.tail->prevNode->data)==0);

  // Now try adding to the head of the list.
  CPPUNIT_ASSERT_MESSAGE("Improper behavior of addindex function.", ll_addIndex(&myList, "CS3841", 7, 0)==true);
  CPPUNIT_ASSERT_MESSAGE("Size improper after add.", myList.size==4);
  CPPUNIT_ASSERT_MESSAGE("Head set up properly.", strcmp("CS3841", (const char*)myList.head->data)==0);
  CPPUNIT_ASSERT_MESSAGE("Head not given null as previous element.", myList.head->prevNode==NULL);
  CPPUNIT_ASSERT_MESSAGE("Next not set up properly from tail add.", strcmp("Hello", (const char*)myList.head->nextNode->data)==0);

  // Now try adding in the middle of the list.
  CPPUNIT_ASSERT_MESSAGE("Improper tail element with addIndex function.", ll_addIndex(&myList, "World", 6, 2)==true);
  CPPUNIT_ASSERT_MESSAGE("Size improper after add.", myList.size==5);

  // Now walk ther list and check everything.
  // List list should be the following.
  
  
 
  CPPUNIT_ASSERT_MESSAGE("Head set up properly.", strcmp("CS3841", (const char*)(myList.head->data))==0);
  CPPUNIT_ASSERT_MESSAGE("Head not given null as previous element.", myList.head->prevNode==NULL);
  CPPUNIT_ASSERT_MESSAGE("Next not set up properly from tail add.", strcmp("Hello", (const char*)myList.head->nextNode->data)==0);
  
  struct listNode* n = myList.head;
  CPPUNIT_ASSERT_MESSAGE("Head is incorrect.", strcmp("CS3841", (const char*)n->data)==0);
  CPPUNIT_ASSERT_MESSAGE("Improper next link.", n->nextNode!=NULL);
  CPPUNIT_ASSERT_MESSAGE("Improper previous link.", n->prevNode==NULL);

  // Walk to the next node.
  n = n->nextNode;
  CPPUNIT_ASSERT_MESSAGE("Node data is incorrect.", strcmp("Hello", (const char*)n->data)==0);
  CPPUNIT_ASSERT_MESSAGE("Improper next link.", n->nextNode!=NULL);
  CPPUNIT_ASSERT_MESSAGE("Improper previous link.", n->prevNode!=NULL);

  // Walk to the next node.
  n = n->nextNode;
  CPPUNIT_ASSERT_MESSAGE("Node data is incorrect.", strcmp("World", (const char*)n->data)==0);
  CPPUNIT_ASSERT_MESSAGE("Improper next link.", n->nextNode!=NULL);
  CPPUNIT_ASSERT_MESSAGE("Improper previous link.", n->prevNode!=NULL);

  // Walk to the next node.
  n = n->nextNode;
  CPPUNIT_ASSERT_MESSAGE("Node data is incorrect.", strcmp("This is London.", (const char*)n->data)==0);
  CPPUNIT_ASSERT_MESSAGE("Improper next link.", n->nextNode!=NULL);
  CPPUNIT_ASSERT_MESSAGE("Improper previous link.", n->prevNode!=NULL);

  // Walk to the next node.
  n = n->nextNode;
  CPPUNIT_ASSERT_MESSAGE("Node data is incorrect.", strcmp("Here is the news.", (const char*)n->data)==0);
  CPPUNIT_ASSERT_MESSAGE("Improper next link.", n->nextNode==NULL);
  CPPUNIT_ASSERT_MESSAGE("Improper previous link.", n->prevNode!=NULL);
}

void LinkedListTestCase::testRemove()
{
  struct mallinfo baselineStatus = mallinfo();

  setupBasicList();

  CPPUNIT_ASSERT_MESSAGE("Checking that a NULL list is ignored.", ll_remove(NULL, 5)==false);
  CPPUNIT_ASSERT_MESSAGE("Checking that an out of range element is ignored.", ll_remove(&myList, 5)==false);


  // Try removing the tail as a first attempt.
  CPPUNIT_ASSERT_MESSAGE("Tail removal failed.", ll_remove(&myList, 4)==true);
  CPPUNIT_ASSERT_MESSAGE("Tail message incorrect after removal.", strcmp("Fall", (const char*)myList.tail->data)==0);
  CPPUNIT_ASSERT_MESSAGE("Tail has a next element.", myList.tail->nextNode==NULL);
  CPPUNIT_ASSERT_MESSAGE("Incorrect size after removal.", myList.size==4);

  // Now try removing the head.
  CPPUNIT_ASSERT_MESSAGE("Tail removal failed.", ll_remove(&myList, 0)==true);
  CPPUNIT_ASSERT_MESSAGE("Tail message incorrect after removal.", strcmp("Operating", (const char*)myList.head->data)==0);
  CPPUNIT_ASSERT_MESSAGE("Head has previous element.", myList.head->prevNode==NULL);
  CPPUNIT_ASSERT_MESSAGE("Incorrect size after removal.", myList.size==3);
  
  // Now try removing from the middle
  CPPUNIT_ASSERT_MESSAGE("Tail removal failed.", ll_remove(&myList, 1)==true);
  CPPUNIT_ASSERT_MESSAGE("Tail message incorrect after removal.", strcmp("Operating", (const char*)myList.head->data)==0);
  CPPUNIT_ASSERT_MESSAGE("Incorrect size after removal.", myList.size==2);

  // Remove from the end of the list again.
  CPPUNIT_ASSERT_MESSAGE("Tail removal failed.", ll_remove(&myList, 1)==true);
  CPPUNIT_ASSERT_MESSAGE("Tail message incorrect after removal.", strcmp("Operating", (const char*)myList.tail->data)==0);
  CPPUNIT_ASSERT_MESSAGE("Tail has a next element.", myList.tail->nextNode==NULL);
  CPPUNIT_ASSERT_MESSAGE("Incorrect size after removal.", myList.size==1);
  CPPUNIT_ASSERT_MESSAGE("Head and tail should match at this point.", myList.head==myList.tail);

  // Now try removing the head.
  CPPUNIT_ASSERT_MESSAGE("Tail removal failed.", ll_remove(&myList, 0)==true);

  // We we should have an empty list.  Check all of the status.
  struct mallinfo endStatus = mallinfo();
  CPPUNIT_ASSERT_MESSAGE("Memory leak in clear operation.", endStatus.uordblks == baselineStatus.uordblks);
  CPPUNIT_ASSERT_MESSAGE("Head not set properly.", myList.head==NULL);
  CPPUNIT_ASSERT_MESSAGE("Tail not set properly.", myList.tail==NULL);
  CPPUNIT_ASSERT_MESSAGE("Size not set properly.", myList.size==0);

}

void LinkedListTestCase::testGet()
{

  const char* messages[] = {"CS3841", "Operating", "Systems", "Fall", "Quarter"};
  int index;

  setupBasicList();
  char *data;

  // Verify basic parameter checks operate properly.
  CPPUNIT_ASSERT_MESSAGE("NULL parameter improperly allowed.", ll_get(NULL, 1)==NULL);
  CPPUNIT_ASSERT_MESSAGE("OUT of range parameter not ignored.", ll_get(&myList, 1000)==false);

  for (index = 0; index < 5; index++)
    {
      data = (char*)ll_get(&myList, index);
      CPPUNIT_ASSERT_MESSAGE("Improper message on get", strcmp(messages[index], data)==0);
    }  
}

void LinkedListTestCase::testClear()
{
  struct mallinfo baselineStatus = mallinfo();

  setupBasicList();
  CPPUNIT_ASSERT_MESSAGE("Starting size incorrect.", ll_size(&myList)==5);
 
  struct mallinfo midlineStatus = mallinfo();
  CPPUNIT_ASSERT_MESSAGE("Memory allocation problem.", midlineStatus.uordblks > baselineStatus.uordblks);

  ll_clear(&myList);
  struct mallinfo endStatus = mallinfo();
  CPPUNIT_ASSERT_MESSAGE("Memory leak in clear operation.", endStatus.uordblks == baselineStatus.uordblks);
  CPPUNIT_ASSERT_MESSAGE("Head not set properly.", myList.head==NULL);
  CPPUNIT_ASSERT_MESSAGE("Tail not set properly.", myList.tail==NULL);
  CPPUNIT_ASSERT_MESSAGE("Size not set properly.", myList.size==0);
}

void LinkedListTestCase::testSize()
{
  CPPUNIT_ASSERT_MESSAGE("Empty list does not have 0 size.", ll_size(&myList)==0);
  CPPUNIT_ASSERT_MESSAGE("Size does not match structure.", ll_size(&myList)==myList.size);

  // Add something and make certain the size goes up.
  ll_add(&myList, "Hello", 6);

  CPPUNIT_ASSERT_MESSAGE("1 element list does not have 1 size.", ll_size(&myList)==1);
  CPPUNIT_ASSERT_MESSAGE("Size does not match structure.", ll_size(&myList)==myList.size);
}

void LinkedListTestCase::testIterator()
{
  const char* messages[] = {"CS3841", "Operating", "Systems", "Fall", "Quarter"};
  int index;

  setupBasicList();
  char* data;

  // Start by testing the obtaining of the iterator.
  struct linkedListIterator *iter;

  iter = ll_getIterator(NULL);
  CPPUNIT_ASSERT_MESSAGE("Iterator should be NULL.", iter==NULL);

  // Now test obtaining a valid iterator.
  iter = ll_getIterator(&myList);
  CPPUNIT_ASSERT_MESSAGE("Iterator should not be NULL.", iter!=NULL);
  CPPUNIT_ASSERT_MESSAGE("Iterator's current should be the head of the list.", iter->current==myList.head);

  for (index = 0; index < 5; index++)
    {
      CPPUNIT_ASSERT_MESSAGE("An invalid iterator should result in a false value.", ll_hasNext(NULL)==false);
      CPPUNIT_ASSERT_MESSAGE("Iterator should have a next right now.", ll_hasNext(iter)==true);


      CPPUNIT_ASSERT_MESSAGE("Invalid parameter to next should result in NULL being returned,", ll_next(NULL)==NULL);
      data = (char*)ll_next(iter);
      CPPUNIT_ASSERT_MESSAGE("Expected true for correct next operation.", data!=NULL);
      CPPUNIT_ASSERT_MESSAGE("Data returned does not match expected data.", strcmp(messages[index], data)==0);
    }  

  // Now we have stepped through everything.  There should be nothing left for the next.
  CPPUNIT_ASSERT_MESSAGE("Extra next operation.", ll_hasNext(iter)==false); 
  CPPUNIT_ASSERT_MESSAGE("Invalid return value when attempting to retrieve beyodn the end of the list.", ll_next(iter)==NULL);
  
  free(iter);

}
